@@ -8,14 +8,12 @@ use Fcntl;
 use IO::Handle;
 use IO::Socket::INET;
 use IO::Socket::UNIX;
-use List::MoreUtils qw(uniq);
 use POSIX qw(:sys_wait_h);
-use Proc::Wait3;
-use Scope::Guard;
+use Server::Starter::Guard;
 
 use Exporter qw(import);
 
-our $VERSION = '0.19';
+our $VERSION = '0.21';
 our @EXPORT_OK = qw(start_server restart_server server_ports);
 
 my @signals_received;
@@ -64,7 +62,7 @@ sub start_server {
             or die "failed to open file:$opts->{pid_file}: $!";
         print $fh "$$\n";
         close $fh;
-        return Scope::Guard->new(
+        return Server::Starter::Guard->new(
             sub {
                 unlink $opts->{pid_file};
             },
@@ -85,7 +83,7 @@ sub start_server {
     }
     
     # create guard that removes the status file
-    my $status_file_guard = $opts->{status_file} && Scope::Guard->new(
+    my $status_file_guard = $opts->{status_file} && Server::Starter::Guard->new(
         sub {
             unlink $opts->{status_file};
         },
@@ -123,7 +121,7 @@ sub start_server {
         push @sockenv, "$port=" . $sock->fileno;
         push @sock, $sock;
     }
-    my $path_remove_guard = Scope::Guard->new(
+    my $path_remove_guard = Server::Starter::Guard->new(
         sub {
             -S $_ and unlink $_
                 for @$paths;
@@ -151,9 +149,9 @@ sub start_server {
     $ENV{SERVER_STARTER_GENERATION} = 0;
     
     # setup signal handlers
-    $SIG{$_} = sub {
+    _set_sighandler($_, sub {
         push @signals_received, $_[0];
-    } for (qw/INT TERM HUP ALRM/);
+    }) for (qw/INT TERM HUP ALRM/);
     $SIG{PIPE} = 'IGNORE';
     
     # setup status monitor
@@ -215,10 +213,10 @@ sub start_server {
         my @r;
         if ($block && $ENV{ENABLE_AUTO_RESTART}) {
             alarm(1);
-            @r = wait3($block);
+            @r = _wait3($block);
             alarm(0);
         } else {
-            @r = wait3($block);
+            @r = _wait3($block);
         }
         return @r;
     };
@@ -234,7 +232,7 @@ sub start_server {
         kill $term_signal, $_
             for sort keys %old_workers;
         while (%old_workers) {
-            if (my @r = wait3(1)) {
+            if (my @r = _wait3(1)) {
                 my ($died_worker, $status) = @r;
                 print STDERR "worker $died_worker died, status:$status\n";
                 delete $old_workers{$died_worker};
@@ -336,7 +334,13 @@ sub restart_server {
     my $get_generations = sub {
         open my $fh, '<', $opts->{status_file}
             or die "failed to open file:$opts->{status_file}:$!";
-        uniq sort { $a <=> $b } map { /^(\d+):/ ? ($1) : () } <$fh>;
+        my %gen;
+        while (my $line = <$fh>) {
+            if ($line =~ /^(\d+):/) {
+                $gen{$1} = 1;
+            }
+        }
+        sort { $a <=> $b } keys %gen;
     };
     
     # wait for this generation
@@ -382,6 +386,40 @@ sub _reload_env {
     return %env;
 }
 
+our $sighandler_should_die;
+my $sighandler_got_sig;
+
+sub _set_sighandler {
+    my ($sig, $proc) = @_;
+    $SIG{$sig} = sub {
+        $proc->(@_);
+        $sighandler_got_sig = 1;
+        die "got signal"
+            if $sighandler_should_die;
+    };
+}
+
+sub _wait3 {
+    my $block = shift;
+    my $pid = -1;
+    if ($block) {
+        local $@;
+        eval {
+            $sighandler_got_sig = 0;
+            local $sighandler_should_die = 1;
+            die "exit from eval"
+                if $sighandler_got_sig;
+            $pid = wait();
+        };
+        if ($pid == -1 && $@) {
+            $! = Errno::EINTR;
+        }
+    } else {
+        my $pid = waitpid(-1, WNOHANG);
+    }
+    return $pid > 0 ? ($pid, $?) : ();
+}
+
 1;
 __END__
 
